from pathlib import Path

from dotenv import load_dotenv


class ConfigError(Exception):
    pass

class AttrDict(dict):

    def __init__(self, d=None, **kwargs):
        super().__init__()
        if d is None:
            d = {}
        else:
            d = dict(d)
        if kwargs:
            d.update(**kwargs)
        for k, v in d.items():
            setattr(self, k, v)
        # Class attributes
        for k in self.__class__.__dict__.keys():
            if (not (k.startswith('__') and k.endswith('__')) and
                    k not in ('update', 'pop', 'from_env', 'from_yaml', 'from_nacos')):
                setattr(self, k, getattr(self, k))

    def __setattr__(self, name, value):
        if isinstance(value, (list, tuple)):
            value = type(value)(self.__class__(x)
                                if isinstance(x, dict) else x for x in value)
        elif isinstance(value, dict) and not isinstance(value, AttrDict):
            value = AttrDict(value)
        super(AttrDict, self).__setattr__(name, value)
        super(AttrDict, self).__setitem__(name, value)

    # 将 __setitem__ 方法直接指向 __setattr__ 方法
    __setitem__ = __setattr__

    def update(self, e=None, **f):
        d = e or dict()
        d.update(f)
        for k in d:
            setattr(self, k, d[k])

    def pop(self, k, *args):
        if hasattr(self, k):
            delattr(self, k)
        return super(AttrDict, self).pop(k, *args)

    @classmethod
    def from_yaml(cls, file_path=None, text=None):
        import yaml
        if file_path:
            file_path = Path(file_path)
            if file_path.exists():
                with open(file_path, 'r') as file:
                    data = yaml.safe_load(file) or {}
        elif text:
            data = yaml.load(text, yaml.FullLoader)
        return cls(data)

    @classmethod
    def from_env(cls, file_path=None):
        load_dotenv(file_path)

    @classmethod
    def from_nacos(cls, env=True, **kwargs):
        """获取NaCos配置
        env: 默认优先加载环境变量
        NACOS_SERVER_ADDRESSES = "localhost:8848"
        NACOS_DATAID = "example-config"
        NACOS_GROUP = "DEFAULT_GROUP"
        """
        import os
        import requests
        if env:
            cls.from_env()
            # 优先加载环境变量
            server_addresses = os.getenv('NACOS_SERVER_ADDRESSES') or kwargs.get('NACOS_SERVER_ADDRESSES')
            data_id = os.getenv('NACOS_DATAID') or kwargs.get('NACOS_DATAID')
            group = os.getenv('NACOS_GROUP') or kwargs.get('NACOS_GROUP')
        else:
            server_addresses = kwargs.get('NACOS_SERVER_ADDRESSES')
            data_id = kwargs.get('NACOS_DATAID')
            group = kwargs.get('NACOS_GROUP')
        if not server_addresses:
            # raise ValueError('NACOS_SERVER 不存在!')
            raise ConfigError('NACOS_SERVER 不存在!')
        config_response = requests.get(
            server_addresses,
            params={"dataId": data_id,
                    "group": group}
        )
        return cls.from_yaml(text=config_response.text)


if __name__ == '__main__':
    # 读取环境变量
    import os
    import requests

    os.environ['NACOS_SERVER_ADDRESSES'] = 'http://10.168.2.83:8848/nacos/v1/cs/configs'
    os.environ['NACOS_DATAID'] = 'py-achievement.yaml'
    os.environ['NACOS_GROUP'] = 'DEFAULT_GROUP'
    config = AttrDict.from_nacos()
    print(config.db)
    print(config.db.host)  # 输出: localhost



# -*- coding: utf-8 -*-
# @time: 2024/2/18 10:32
# @author: Dyz
# @file: schemas.py
# @software: PyCharm
import json
import re
from typing import List

from pydantic import BaseModel, validator


def reset_issn(issn: str):
    if issn:
        issn = issn.upper()
    if '*' in issn or 'N/A' in issn:
        return ''
    if len(issn) > 9:
        issn = re.findall('([0-9A-Z]{4}-[0-9A-Z]{4})', issn)
        if issn:
            issn = issn[0]
        else:
            issn = ''
    return issn.upper()


class journalSN(BaseModel):
    """期刊 ISSN 处理"""
    issn: str = ''
    eissn: str = ''

    @validator('issn')
    def check_issn(cls, val):
        """检查issn"""
        return reset_issn(val)

    @validator('eissn')
    def check_eissn(cls, val):
        """检查issn"""
        return reset_issn(val)

    def get_oissn(self) -> List:
        sn_set = set()
        if self.issn:
            sn_set.add(self.issn)
        if self.eissn:
            sn_set.add(self.eissn)
        return list(sn_set)


class BaseSchemas(BaseModel):
    """Base Schemas"""

    @staticmethod
    def dump(val):
        if isinstance(val, (list, dict)):
            return json.dumps(val, ensure_ascii=False)
        return val

    def value_dump(self) -> dict:
        """转换值，便于存储数据库"""
        return {k: self.dump(v) for k, v in self.dict().items()}



# -*- coding: utf-8 -*-
# @time: 2022/4/21 16:50
# @author: Dyz
# @file: time.py
# @software: PyCharm
import asyncio
from functools import wraps
from pathlib import Path
from time import gmtime, strftime, time


def get_args_file(*args, **kwargs):
    """从传入参数中找寻file相关的参数"""
    for file in args:
        if isinstance(file, Path):
            return file
    for key, val in kwargs.items():
        if isinstance(val, Path):
            return val


def _timer(func, start_time, *args, **kwargs):
    """计时与打印"""
    time_ = strftime("%H:%M:%S", gmtime(time() - start_time))
    if 'download' in str(func.__name__):
        file = get_args_file(*args, **kwargs)
        print(f'{file or str(func.__name__)} 耗时: {time_}')
    else:
        print(f'{func.__name__} 耗时: {time_}')


def timer(func):
    """ 计算时间装饰器 """
    if asyncio.iscoroutinefunction(func):
        # 用于装饰异步函数
        @wraps(func)
        async def aio_wrapper(*args, **kwargs):
            start_time = time()
            res = await func(*args, **kwargs)
            _timer(func, start_time, *args, **kwargs)
            return res

        return aio_wrapper

    else:
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time()
            res = func(*args, **kwargs)
            _timer(func, start_time, *args, **kwargs)
            return res

        return wrapper



# -*- coding: utf-8 -*-
# @time: 2024/2/18 9:49
# @author: Dyz
# @file: base_conn.py
# @software: PyCharm
import functools
from typing import Optional

from loguru import logger
from tortoise import BaseDBAsyncClient, models, fields, Tortoise


async def init_db(tortoise_orm: dict, create: bool = False):
    """
    >>> {
        "connections": {
            "conn": {
                "engine": f"tortoise.backends.mysql",  # mysql/asyncpg/sqlite
                "credentials": {
                    'host': 'localhost',
                    'port': 3306,
                    'user': 'root',
                    'password': 'xxx',
                    'database': 'db_name',
                    'charset': 'utf8mb4'
                }
            }
        },
        "apps": {
            "jour": {"models": ["__main__"], "default_connection": "conn"},
        },
        'use_tz': False,
        'timezone': 'Asia/Shanghai',
    }
    """
    await Tortoise.init(
        config=tortoise_orm
    )
    if create:
        await Tortoise.generate_schemas()


async def close_db():
    """关闭数据库连接"""
    await Tortoise.close_connections()


def load_db(tortoise_orm, create: bool = False):
    """
    tortoise_orm: 数据库链接配置
    create: 创建表
    """

    def _load_db(func):
        """加载数据库"""

        @functools.wraps(func)
        async def wrap(*args, **kwargs):
            await init_db(tortoise_orm, create)

            result = await func(*args, **kwargs)

            await close_db()
            return result

        return wrap

    return _load_db


class DataModel(models.Model):
    """提供 数据截断表 与 复制的快捷操作"""
    id = fields.IntField(pk=True)

    class Meta:
        abstract = True  # 标记为抽象类，不直接创建表

    @classmethod
    async def truncate_model(cls, using_db: Optional[BaseDBAsyncClient] = None):
        """截断表"""
        db = using_db or cls._choose_db(True)
        db_name = cls._meta.db_table
        if db.capabilities.dialect == 'postgres':  # mysql
            await db.execute_script(f'TRUNCATE TABLE {db_name} RESTART IDENTITY')
            logger.info(f'截断表[{db_name}]')
        else:
            await db.execute_script(f'TRUNCATE TABLE {db_name}')
            logger.info(f'截断表[{db_name}]')

    @classmethod
    async def copy_old(cls, using_db: Optional[BaseDBAsyncClient] = None, name='_old'):
        """复制表"""
        db = using_db or cls._choose_db(True)
        _table = cls._meta.db_table
        old_table = _table + name
        data = await cls.all().limit(1)
        if data:
            await db.execute_query(f'DROP TABLE IF EXISTS {old_table}')
            return await db.execute_query(f'CREATE TABLE {old_table} SELECT * FROM {_table}')


class BaseModel(DataModel):
    """基础模型类，包含创建时间和更新时间"""

    class Meta:
        abstract = True  # 标记为抽象类，不直接创建表

    created_at = fields.DatetimeField(auto_now_add=True)
    updated_at = fields.DatetimeField(auto_now=True)

# class Journal(BaseModel):
#     class Meta:
#         table = 'journal' # 表名
#         unique_together = (("uid", "uid2", "word"),) # 联合唯一索引
#     table_description = '主题词查重数据表'  # 表描述



import json
import os
from hashlib import md5

from Crypto import Random
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_v1_5 as PKCS1_cipher
import base64


def make_md5(s: str, encoding='utf-8') -> str:
    """加密"""
    return md5(s.encode(encoding)).hexdigest()


class RsaCrypto:
    def __init__(self, b=1024, dir='./'):
        random_generator = Random.new().read
        self.rsa = RSA.generate(b, random_generator)
        self.dir = dir

    def create_rsa_key(self):
        """生成RSA秘钥对"""
        public_key = self.rsa.publickey().exportKey()  # 公钥
        private_key = self.rsa.exportKey()  # 私钥
        with open(os.path.join(self.dir, "id_rsa.pub"), 'wb') as f2:
            f2.write(public_key)
        with open(os.path.join(self.dir, "id_rsa"), 'wb') as f:
            f.write(private_key)

    def encrypt(self, text, public_key=None):
        """加密"""
        if not public_key:
            with open(os.path.join(self.dir, 'id_rsa.pub')) as f:
                public_key = f.read()
        pub_key = RSA.importKey(str(public_key))
        cipher = PKCS1_cipher.new(pub_key)
        rsa_text = base64.b64encode(cipher.encrypt(bytes(text.encode("utf8"))))
        # print(rsa_text.decode('utf-8'))
        return rsa_text.decode('utf-8')

    def decrypt(self, text_data, private_key=None):
        """解密"""
        if not private_key:
            with open(os.path.join(self.dir, 'id_rsa')) as f:
                private_key = f.read()
        pri_key = RSA.importKey(private_key)
        cipher = PKCS1_cipher.new(pri_key)
        back_text = cipher.decrypt(base64.b64decode(text_data), 0)
        # print(back_text.decode('utf-8'))
        return back_text.decode('utf-8')


r = RsaCrypto()


def create_rsa_key():
    """加密"""
    return r.create_rsa_key()


def encrypt(text):
    """加密"""
    return r.encrypt(text)


def decrypt(text):
    """解密"""
    return r.decrypt(text)


if __name__ == "__main__":
    data = {"user": "1959474", "name": "H, DJ"}
    a = create_rsa_key()
    a = encrypt(json.dumps(data))
    b = decrypt(a)
    print(a)
    print(json.loads(b))



# -*- coding: utf-8 -*-
# @time: 2024/3/7 15:21
# @author: Dyz
# @file: Iterator_xml.py
# @software: PyCharm
# 读取XML 文件
from pathlib import Path


def iterator_read(file: Path, start: str, encoding='utf-8'):
    """
    迭代器读取
    """
    with open(file, encoding=encoding) as f:
        row_xml = ''
        for line in f:
            # 每条数据的开头
            if line.startswith(start):
                if row_xml:  # 如果有上一条就返回后赋予新的
                    yield row_xml
                    row_xml = line
                    continue
                else:  # 没有上一条就直接加在一起
                    row_xml += line
                    continue
            if row_xml:
                row_xml += line
        if row_xml:
            yield row_xml



# -*- coding: utf-8 -*-
# @time: 2024/2/18 9:39
# @author: Dyz
# @file: base_spider.py
# @software: PyCharm
import asyncio
import logging
from pathlib import Path

import httpx
from requests import Session
from fake_useragent import UserAgent
from tenacity import retry, stop_after_attempt, wait_fixed

from du_aio_tools.exceptions import HTTPError
from du_aio_tools.time import timer

logger = logging.getLogger("du_aio_tools.BaseSpider")

MAX_RETRY = 3  # 最大重试次数
WAIT_FIXED = 3  # 重试间隔 3s
TIMEOUT = 10  # 重试间隔 10s


class BaseDataSpider:

    @staticmethod
    async def save_db(table, data):
        """ 保存数据库 """
        await table.create(**data)

    @staticmethod
    async def bulk_save_db(table, data_list):
        """
        批量插入:  Iterable[MODEL], 列表中的元素应为 model
        """
        await table.bulk_create(data_list)

    @staticmethod
    async def parse(*args, **kwargs):
        """
        解析HTML数据 parsel 是 scrapy 内置的解析器
            >>> from parsel import Selector
            >>> doc = Selector(text=resp.text)
        解析json数据
            >>> data = resp.json()
        """
        pass


class BaseSpider(BaseDataSpider):
    """基础爬虫"""

    def __init__(self, name=None,
                 start_url=None,
                 headers=None,
                 ):
        """
        name: str 爬虫名称/域名/表名， 用于记录日志
        """
        self.name = name
        self.start_url = start_url
        self.session: Session = Session()
        self.headers = headers or {}
        self.last_ua = ''
        self.ua = UserAgent()

    def random_ua(self, headers=None):
        """随机 UA"""
        if headers:
            headers['User-Agent'] = self.get_ua()
            return headers
        else:
            self.headers['User-Agent'] = self.get_ua()
            return self.headers

    def get_ua(self):
        while True:
            ua_ = self.ua.random
            # 不是IE 并且 不是上一次使用的 UA
            if 'MSIE' not in str(ua_) and self.last_ua != ua_:
                self.last_ua = ua_
                return ua_

    def responses(self, resp, **kwargs):
        """可以再次处理相关异常"""
        if resp.status_code >= 400:
            # 请求异常
            self.random_ua(**kwargs)  # 切换 UA 重试时使用新UA
            raise HTTPError(f'{resp.url}-{resp.status_code}请求异常')
        return resp

    @retry(retry=stop_after_attempt(MAX_RETRY), wait=wait_fixed(WAIT_FIXED))
    def get(self, url, **kwargs):
        """get 请求"""
        resp = self.session.get(url, **kwargs)
        return self.responses(resp)

    @retry(retry=stop_after_attempt(MAX_RETRY), wait=wait_fixed(WAIT_FIXED))
    def post(self, url, **kwargs):
        """post 请求"""
        resp = self.session.post(url, **kwargs)
        return self.responses(resp)

    @staticmethod
    def parse_range(file_size, slice):
        arr = list(range(0, file_size, slice))
        result = []
        for _ in range(len(arr) - 1):
            s_pos, e_pos = arr[_], arr[_ + 1] - 1
            result.append([s_pos, e_pos])
        result[-1][-1] = file_size - 1
        return result

    @timer
    def download(self, url, file: Path, method='get', **kwargs):
        """requests 数据流 方式下载文件"""
        if not file.exists():
            file.touch()  # 创建文件

        if method.lower() == 'get':  # 1M
            res = self.get(url, **kwargs, stream=True)
            with file.open('wb') as f:
                for chunk in res.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)
        else:
            res = self.post(url, **kwargs, stream=True)
            with file.open('wb') as f:
                for chunk in res.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)


class AioSpider(BaseSpider):
    def __init__(self, name=None, headers=None, conn_pool: bool = True, proxy=None, sem: int = 15):
        """不使用连接池的例子， 如使用代理, 需要传入特殊参数 client_kwargs
        >> url = 'https://myip.ipip.net/'
        >> aio_spider = AioSpider(conn_pool=False)
        >> proxy_url = "http://%(user)s:%(pwd)s@%(proxy)s/" % {"user": username,
        >>                                                     "pwd": password,
        >>                                                     "proxy": tunnel}

        >> res = await aio_spider.aio_get('https://myip.ipip.net/')
        >> print(res.text)

        name: str 爬虫名称/域名/表名
        headers: 请求头参数
        conn_pool: bool 是否使用全局连接池, 部分网站使用ip代理会在使用 httpx 连接池时异常
        sem: 连接池设置并发
        """
        self.proxy = proxy
        super().__init__(name=name, headers=headers)
        self.conn_pool = conn_pool
        self.client: httpx.AsyncClient | None = None
        if conn_pool:
            # max_keepalive_connections 允许的保持活动状态连接数，或None始终允许。（默认值 20）
            # max_connections 允许的最大连接数，或None无限制。（默认值 100）
            limits = httpx.Limits(max_keepalive_connections=sem, max_connections=sem + 3)
            self.client = httpx.AsyncClient(limits=limits, proxy=self.proxy)  # 异步

    @retry(retry=stop_after_attempt(MAX_RETRY), wait=wait_fixed(WAIT_FIXED))
    async def aio_get(self, url, **kwargs):
        """get 请求"""
        if self.conn_pool:
            resp = await self.client.request("GET", url, follow_redirects=True, **kwargs)
        else:
            async with httpx.AsyncClient(proxy=self.proxy) as client:
                resp = await client.request("GET", url, follow_redirects=True, **kwargs)

        return self.responses(resp)

    @retry(retry=stop_after_attempt(MAX_RETRY), wait=wait_fixed(WAIT_FIXED))
    async def aio_post(self, url, **kwargs):
        """post 请求"""
        if self.conn_pool:
            resp = await self.client.request("POST", url, follow_redirects=True, **kwargs)
        else:
            async with httpx.AsyncClient(proxy=self.proxy) as client:
                resp = await client.request("POST", url, follow_redirects=True, **kwargs)
        return self.responses(resp)

    @retry(retry=stop_after_attempt(MAX_RETRY), wait=wait_fixed(WAIT_FIXED))
    async def async_range_download(self, url, file: Path, s_pos, e_pos, method='get', **kwargs):
        """异步下载 seek 写入对应位置"""
        headers = {"Range": f"bytes={s_pos}-{e_pos}"}
        if method.lower() == 'get':
            res = await self.aio_get(url, headers=headers, **kwargs)
        else:
            res = await self.aio_post(url, headers=headers, **kwargs)
        with open(file, "rb+") as f:
            f.seek(s_pos)
            f.write(res.content)

    @timer
    async def aio_download(self, url, file: Path, method='get', slice=5, **kwargs):
        """httpx 异步分片式 下载文件
            slice 默认最高分成10个分片， 同时并发下载
        """
        if not file.exists():
            file.touch()  # 创建文件  片

        res = httpx.head(url, **kwargs)
        filesize = int(res.headers['Content-Length'])
        if filesize > 1_000_000:  # 1M
            divisional_ranges = self.parse_range(file_size=filesize, slice=slice)
            tasks = [self.async_range_download(url, file, s_pos, e_pos, method=method) for s_pos, e_pos in
                     divisional_ranges]
            await asyncio.gather(*tasks)  # 异步并发下载
        else:
            if method.lower() == 'get':  # 1M
                res = await self.aio_get(url, **kwargs)
            else:
                res = await self.aio_post(url, **kwargs)
            file.write_bytes(res.content)


class BrowserDriver(BaseSpider):
    """浏览器驱动 爬虫"""

    def __init__(self, context, name=None):
        super().__init__(name)
        self.context = context

    @staticmethod
    async def add_js(page, js=None):
        js = js or """Object.defineProperties(navigator, {webdriver:{get:()=>undefined}});"""
        await page.add_init_script(js)

    async def new_page(self):
        page = await self.context.new_page()
        await self.add_js(page)
        return page

    @staticmethod
    async def get_html(page):
        """获取page 的 网页源码"""
        return await page.content()

    @staticmethod
    async def download_file(url, page, file: Path, click=None):
        """下载文件"""
        await page.goto(url)
        async with page.expect_download() as download2_info:
            if click:
                await page.locator(click).first.click()
        download = await download2_info.value
        await download.save_as(file)


async def example():
    """浏览器例子"""
    remote_ip = 'ws://127.0.0.1:3000'
    from playwright.async_api import async_playwright
    async with async_playwright() as p:
        if remote_ip:
            browser = await p.chromium.connect_over_cdp(remote_ip)
        else:
            browser = await p.chromium.launch(headless=False)
        context = await browser.new_context()

        br = BrowserDriver(context=context, name='测试网站-浏览器爬虫')
        page = await br.new_page()
        await page.goto('https://earlywarning.fenqubiao.com/#/zh-cn/early-warning-journal-list-2024')
        pass


if __name__ == '__main__':
    asyncio.run(example())
